#include <sys/types.h>
#include <netinet/in.h>
#include <unistd.h>
#include <stdio.h>
#include <errno.h>
#include <signal.h>
#include <netdb.h>
#include "PsxIO.h"

int PsxRead (int fd, char *buf, long nbyte, int elsize)
{
  long bytes_so_far    = 0;
  long bytes_this_time = 0;
  long bytes_left      = nbyte;
  long cswap;

  while (bytes_left) {
    bytes_this_time = read (fd, buf,
			    (bytes_left>1024) ? 1024 : bytes_left);
    
    if (bytes_this_time < 0) {
      if (errno == EINTR)
	continue;
      return (bytes_this_time);
    }
    else {
      if (elsize==sizeof(long) && (1 != htonl (1))) {
	for (cswap=0; cswap<bytes_this_time;
	     cswap+=sizeof(long), buf+=sizeof(long))
	  *((long *) buf) = htonl (*((long *) buf));
      }
      else if (elsize==sizeof(short) && (1 != htons (1))) {
	for (cswap=0; cswap<bytes_this_time;
	     cswap+=sizeof(short),buf+=sizeof(short))
	  *((short *) buf) = htons (*((short *) buf));
      }
      else {
	buf        += bytes_this_time;
      }
      bytes_so_far += bytes_this_time;
      bytes_left   -= bytes_this_time;
    }
  }
  return (bytes_so_far);
} 

int PsxWrite (int fd, char *buf, long nbyte, int elsize)
{
  static char sendbuf[1024], *sbptr;
  long bytes_so_far    = 0;
  long bytes_this_time = 0;
  long sb_so_far       = 0;
  long bytes_left      = nbyte;
  long cswap, num;
  
  if ((elsize==sizeof(long)  && (1 != htonl (1))) ||
      (elsize==sizeof(short) && (1 != htons (1)))) {
    /*
     * wrong byteorder
     */
    while (bytes_left) {
      memcpy (sbptr=sendbuf, buf, num=((bytes_left>1024)?1024:bytes_left));

      if (elsize==sizeof(long) && (1 != htonl (1))) {
	for (cswap=0; cswap<num; cswap+=sizeof(long),sbptr+=sizeof(long))
	  *((long *) sbptr) = htonl (*((long *) sbptr));
      }
      else {
	for (cswap=0; cswap<num; cswap+=sizeof(short),sbptr+=sizeof(short))
	  *((short *) sbptr) = htons (*((short *) sbptr));
      }
      sb_so_far = 0;
      sbptr     = sendbuf;

      while (sb_so_far < num) {
	bytes_this_time = write (fd, sbptr, num - sb_so_far);

	if (bytes_this_time < 0) {
	  if (errno == EINTR)
	    continue;
	  return (bytes_this_time);
	}
	sb_so_far += bytes_this_time;
	sbptr     += bytes_this_time;
      }
      buf          += num;
      bytes_so_far += num;
      bytes_left   -= num;
    }
  }
  else {
    /*
     * correct (network) byteorder
     */
    while (bytes_left) {
      bytes_this_time = write (fd, buf, (bytes_left>1024)?1024:bytes_left);

      if (bytes_this_time < 0) {
	if (errno == EINTR)
	  continue;
	return (bytes_this_time);
      }
      else {
	bytes_so_far += bytes_this_time;
	buf          += bytes_this_time;
	bytes_left   -= bytes_this_time;
      }
    }
  }
  return (bytes_so_far);
}

