#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <stdio.h>
#include <time.h>
#include <stdlib.h>
#include <unistd.h>
#include <signal.h>
#include <netdb.h>
#include <string.h>
#include <signal.h>
#include <errno.h>
#include "PsxIO.h"
#include "comm.h"
#include "bigint.h"

#define MAXSLAVES	256
#define NUMBUCKETS	8

/*
 * Grrrr ... Why, Linux, Why?
 */

#ifndef FD_ZERO
#ifndef __FD_ZERO
# include <gnu/types.h>
#endif
#define FD_ZERO		__FD_ZERO
#define FD_SET		__FD_SET
#define FD_CLR		__FD_CLR
#define FD_ISSET	__FD_ISSET
#endif


typedef struct {
  int    active;
  int    fildes;
  struct sockaddr_in addr;

  long spdidx;

  int  status;
  int  aufgabe;
  int  bucket;
  int  shift;

  long anfang;
  long ende;
  long total;
  long mults;

  time_t calctime;
  time_t waittime;
  time_t transtime;
} procstat;

typedef struct {
    int inuse;
    int numprocs;	/* number of results to collect */
    int aktprocs;	/* number of results already collected */
    bigint value;
} mulbuck;


/*
 * global variables
 */

int       thesock;
int       procs;
procstat *pstat;
mulbuck  *bucket;
bigint    result;
long      theend, sofar, step;
time_t    mtime=0;
long      spiref=0;

int       domult=0;
int       mult1f=0;
int       mult2f=0;
bigint    mult1;
bigint    mult2;
bigint    rectemp;

#define AKTTIME(s)	((s)?((long) (time(NULL)-(s))):0)

#define SEND(o,x,l,e,c) { \
			int xyz; \
			xyz = PsxWrite (o, (char *) (x), l, e); \
			if (xyz != l) { \
			  fprintf (stderr, "%8ld Fatal error master(%d): %ld supposed %ld errno=%d\n", \
				   AKTTIME(mtime), __LINE__, (long)xyz, (long)l, errno); \
			  c; \
			} \
		  };

#define RECV(i,x,l,e,c) { \
		        int xyz; \
			xyz = PsxRead (i, (char *) (x), l, e); \
			if (xyz != l) { \
			  fprintf (stderr, "%8ld Fatal error master(%d): %ld supposed %ld errno=%d\n", \
				   AKTTIME(mtime), __LINE__, (long)xyz, (long)l, errno); \
			  c; \
			} \
		  };

/*
 * Init global variables
 */

void init (void)
{
  struct sockaddr_in theaddr;
  int index;

  if ((pstat = (procstat *) malloc (MAXSLAVES * sizeof (procstat))) == NULL) {
    fprintf (stderr, "ERROR! Couldn't allocate resources.\n");
    exit    (1);
  }
  if ((bucket = (mulbuck *) malloc (NUMBUCKETS * sizeof (mulbuck))) == NULL) {
    fprintf (stderr, "ERROR! Couldn't allocate resources.\n");
    free    (pstat);
    exit    (1);
  }
  
  for (index=0; index<MAXSLAVES; index++) {
    pstat[index].active   = 0;
    pstat[index].fildes   = 0;
    pstat[index].status   = UNKNOWN;
    pstat[index].aufgabe  = 0;
    pstat[index].anfang   = 0;
    pstat[index].bucket   =-1;
    pstat[index].shift    = 0;
    pstat[index].ende     = 0;
    pstat[index].total    = 0;
    pstat[index].mults    = 0;
    pstat[index].calctime = 0;
    pstat[index].waittime = 0;
    pstat[index].transtime= 0;
  }
  for (index=0; index<NUMBUCKETS; index++) {
    bi_const (&bucket[index].value);
    bucket[index].inuse   = 0;
  }
  
  bi_const (&result);
  bi_const (&mult1);
  bi_const (&mult2);
  bi_const (&rectemp);

  procs  = 0;
  sofar  = 0;
  mult1f = 0;
  mult2f = 0;

  if ((thesock = socket (AF_INET, SOCK_STREAM, 0)) == -1) {
    fprintf (stderr, "Couldn't create socket (%d)\n", errno);
    free    (bucket);
    free    (pstat);
    exit    (1);
  }
  theaddr.sin_family      = AF_INET;
  theaddr.sin_port        = htons (PORTNUMBER);
  theaddr.sin_addr.s_addr = htonl (INADDR_ANY);

  if (bind (thesock, (struct sockaddr *) &theaddr, sizeof (struct sockaddr_in)) != 0) {
    fprintf (stderr, "Couldn't bind to socket (%d)\n", errno);
    close   (thesock);
    free    (bucket);
    free    (pstat);
    exit    (1);
  }
  if (listen (thesock, 16) != 0) {
    fprintf (stderr, "Couldn't listen to socket (%d)\n", errno);
    close   (thesock);
    free    (bucket);
    free    (pstat);
    exit    (1);
  }
}

void finish (void)
{
  comm mycomm;
  int index;

  for (index=0; index<MAXSLAVES; index++) {
    if (pstat[index].active) {
      mycomm.command = QUIT;
      SEND  (pstat[index].fildes, &mycomm, sizeof (comm), sizeof (long), ;);
      close (pstat[index].fildes);
    }
  }
  for (index=0; index<NUMBUCKETS; index++) {
    bi_dest (&bucket[index].value);
    bucket[index].inuse   = 0;
  }
  close (thesock);
  free  (bucket);
  free  (pstat);
}

void quitonsig (int signal)
{
  finish ();
  exit   (1);
}

int event (void)
{
  int index, res, max=thesock, len;
  fd_set mask;
  long oaddr;

  FD_ZERO (&mask);

  for (index=0; index<MAXSLAVES; index++) {
    if (pstat[index].active && pstat[index].status != DEAD) {
      if (pstat[index].fildes > max)
	max = pstat[index].fildes;
      FD_SET (pstat[index].fildes, &mask);
    }
  }
  FD_SET (thesock, &mask);

 evagain:
  res = select (max+1, (int *) &mask, NULL, NULL, NULL);

  if (res <= 0) {
    if (errno == EINTR)
      goto evagain;
    fprintf (stderr, "%8ld Fatal error master(%d): couldn't select %d\n",
	     AKTTIME(mtime), __LINE__, errno);
    finish  ();
    exit    (1);
  }
  if (mtime == 0)
    mtime = time (NULL);

  if (FD_ISSET (thesock, &mask)) { /* new connection */
    for (index=0; index<MAXSLAVES; index++)
      if (pstat[index].status == UNKNOWN)
	break;

    if (index != MAXSLAVES) {
      len = sizeof (struct sockaddr_in);
      if ((pstat[index].fildes = accept (thesock, (struct sockaddr *) &pstat[index].addr, &len)) == -1) {
	fprintf (stderr, "%8ld Fatal error master(%d): couldn't accept %d\n",
		 AKTTIME(mtime), __LINE__, errno);
	goto evagain;
      }

      oaddr = ntohl (pstat[index].addr.sin_addr.s_addr);

      fprintf (stderr, "%8ld Got new slave %d at %d.%d.%d.%d\n",
	       AKTTIME(mtime),
	       index,
	       (int) ((oaddr >> 24) & 0xff),
	       (int) ((oaddr >> 16) & 0xff),
	       (int) ((oaddr >>  8) & 0xff),
	       (int) ((oaddr      ) & 0xff));

      pstat[index].active   = 1;
      pstat[index].status   = IDLE;
      pstat[index].spdidx   = 1000;
      pstat[index].aufgabe  = 0;
      pstat[index].anfang   = 0;
      pstat[index].bucket   =-1;
      pstat[index].shift    = 0;
      pstat[index].ende     = 0;
      pstat[index].total    = 0;
      pstat[index].mults    = 0;
      pstat[index].calctime = 0;
      pstat[index].waittime = 0;
      pstat[index].transtime= 0;

      return index;
    }
  }
  for (index=0; index<MAXSLAVES; index++) {
    if (pstat[index].active && FD_ISSET (pstat[index].fildes, &mask))
      return index;
  }
  fprintf (stderr, "%8ld Error in master(%d): what happened? (%d)\n",
	   AKTTIME(mtime), __LINE__, errno);

  goto evagain;
}

int sendaufgabe (int prozessor, long anfang, long ende)
{
  comm mycomm;
  
  fprintf (stderr, "%8ld Calculating   %6ld to %6ld  on  %02d\n",
	   AKTTIME(mtime), anfang, ende, prozessor);
  
  mycomm.prozessor      = prozessor;
  mycomm.command        = AUFGABE;
  mycomm.reply          = UNKNOWN;
  mycomm.aufgabe.anfang = anfang;
  mycomm.aufgabe.ende   = ende;
  mycomm.multiply.bucket= -1;
  mycomm.multiply.shift =  0;
  pstat[mycomm.prozessor].status  = BUSY;
  pstat[mycomm.prozessor].aufgabe = AUFGABE;
  pstat[mycomm.prozessor].bucket  = -1;
  pstat[mycomm.prozessor].shift   =  0;
  pstat[mycomm.prozessor].anfang  = mycomm.aufgabe.anfang;
  pstat[mycomm.prozessor].ende    = mycomm.aufgabe.ende;
  
  SEND (pstat[mycomm.prozessor].fildes,
	&mycomm, sizeof (comm), sizeof (long), return 1);

  return 0;
}

int sendmultcom (int prozessor, bigint *z1, bigint *z2)
{
  comm mycomm;

  fprintf (stderr, "%8ld Sending multiply  %6ldx%-6ld to  %02d\n",
	   AKTTIME(mtime), z1->length, z2->length, prozessor);

  mycomm.prozessor        = prozessor;
  mycomm.command          = MULT;
  mycomm.reply            = UNKNOWN;
  mycomm.multiply.bucket  = -1;
  mycomm.multiply.shift   = z1->first + z2->first;
  mycomm.multiply.length1 = z1->length - z1->first + 1;
  mycomm.multiply.length2 = z2->length - z2->first + 1;
  mycomm.multiply.first1  = 0;
  mycomm.multiply.first2  = 0;
    
  pstat[mycomm.prozessor].status  = BUSY;
  pstat[mycomm.prozessor].aufgabe = MULT;
  pstat[mycomm.prozessor].bucket  = mycomm.multiply.bucket;
  pstat[mycomm.prozessor].shift   = mycomm.multiply.shift;

  SEND (pstat[mycomm.prozessor].fildes, 
	&mycomm, sizeof (comm), sizeof (long), return 1);
  SEND (pstat[mycomm.prozessor].fildes,
	z1->digits + z1->first,
	mycomm.multiply.length1 * sizeof (short),
	sizeof (short),
	return 1);
  SEND (pstat[mycomm.prozessor].fildes,
	z2->digits + z2->first,
	mycomm.multiply.length2 * sizeof (short),
	sizeof (short),
	return 1);
  
  return 0;
}

int OLDdistmult (int mb, int d1, int d2, bigint *z1, bigint *z2)
{
  static int iprocs[MAXSLAVES];
  int index, iindex, iter1, iter2, fail;
  long len1, len2, totspeed;
  comm mycomm;

  for (index=0,iindex=0; index<MAXSLAVES; index++) {
    if (pstat[index].active && pstat[index].status == IDLE)
      iprocs[iindex++] = index;
  }
  if (iindex < d1*d2)
    return -1;
  if (bucket[mb].inuse)
    return -1;

  fprintf (stderr, "%8ld Sending multiply  %6ldx%-6ld to Bucket %02d (%02d) [",
	   AKTTIME(mtime), z1->length, z2->length, mb, d1*d2);
  for (index=0, totspeed=0; index<d1*d2; index++) {
    totspeed += pstat[iprocs[index]].spdidx;
    fprintf (stderr, "%02d/%04ld%c",
	     iprocs[index], pstat[iprocs[index]].spdidx, 
	     (index==d1*d2-1)?']':',');
  }
  fprintf (stderr, "\n");

  bucket[mb].inuse    = 1;
  bucket[mb].numprocs = d1*d2;
  bucket[mb].aktprocs = 0;

  iindex = 0;
  for (iter1=d1-1; iter1>=0; iter1--) {
    for (iter2=d2-1; iter2>=0; iter2--) {
      len1 = (z1->length - z1->first) / d1 + 1;
      len2 = (z2->length - z2->first) / d2 + 1;

      mycomm.prozessor        = iprocs[iindex++];
      mycomm.command          = MULT;
      mycomm.reply            = UNKNOWN;
      mycomm.multiply.bucket  = mb;
      mycomm.multiply.shift   = iter1 * len1 + iter2 * len2 + z1->first + z2->first;
      mycomm.multiply.length1 = (iter1==d1-1)?(z1->length-z1->first-((d1-1)*len1)+1):len1;
      mycomm.multiply.length2 = (iter2==d2-1)?(z2->length-z2->first-((d2-1)*len2)+1):len2;
      mycomm.multiply.first1  = 0;
      mycomm.multiply.first2  = 0;

      pstat[mycomm.prozessor].status  = BUSY;
      pstat[mycomm.prozessor].aufgabe = MULT;
      pstat[mycomm.prozessor].bucket  = mycomm.multiply.bucket;
      pstat[mycomm.prozessor].shift   = mycomm.multiply.shift;

      fail = 0;
      SEND (pstat[mycomm.prozessor].fildes, 
	    &mycomm, sizeof (comm), sizeof (long), fail=1);

      if (fail) {
	/* Do something! */
      }

      SEND (pstat[mycomm.prozessor].fildes,
	    z1->digits + z1->first + iter1 * len1,
	    mycomm.multiply.length1 * sizeof (short),
	    sizeof (short),
	    return 1);
      SEND (pstat[mycomm.prozessor].fildes,
	    z2->digits + z2->first + iter2 * len2,
	    mycomm.multiply.length2 * sizeof (short),
	    sizeof (short),
	    return 1);
    }
  }
  /*
   * Hope this was it
   */
  return 0;
}

/*
 * Why should we need the math lib?
 */

long isqrt (long val)
{
  int left=1, right=val/2, res;

  if (val <= 0)
    return 0;
  if (val <= 3)
    right++;

  while (left != right) {
    res = ((left+right)/2)*((left+right)/2);

    if (res < val) {
      if (right-left == 1) {
	if (right*right-val < val-left*left)
	  return right;
	else
	  return left;
      }
      left  = (left+right)/2;
    }
    else if (res > val)
      right = (left+right)/2;
    else if (res == val)
      return ((left+right)/2);
  }
  return left;
}

/*
 * Send a distributed mult with respect to each process' speed.
 * We rely on that z1 is the larger number, and that it its size
 * is sufficient for a good distribution
 */

int distmult (int mb, int num, bigint *z1, bigint *z2)
{
  static int iprocs[MAXSLAVES];
  int index, iindex, fail;
  long len1, totspeed, first1;
  comm mycomm;

  for (index=0,iindex=0; index<MAXSLAVES; index++) {
    if (pstat[index].active && pstat[index].status == IDLE && pstat[index].spdidx > 10)
      iprocs[iindex++] = index;
  }

  if (iindex < num)
    return -1;
  if (bucket[mb].inuse)
    return -1;


  fprintf (stderr, "%8ld Sending multiply  %6ldx%-6ld to Bucket %02d (%02d) [",
	   AKTTIME(mtime), z1->length, z2->length, mb, num);
  for (index=0, totspeed=0; index<num; index++) {
    totspeed += pstat[iprocs[index]].spdidx;
    fprintf (stderr, "%02d%c",
	     iprocs[index],
	     (index==num-1)?']':',');
  }
  fprintf (stderr, "\n");

  bucket[mb].inuse    = 1;
  bucket[mb].numprocs = num;
  bucket[mb].aktprocs = 0;

  for (iindex=0, first1=z1->first; iindex < num; iindex++) {
    len1 = (((z1->length - first1 + 1) * 
	     pstat[iprocs[iindex]].spdidx) /
	    totspeed);

    mycomm.prozessor        = iprocs[iindex];
    mycomm.command          = MULT;
    mycomm.reply            = UNKNOWN;
    mycomm.multiply.bucket  = mb;
    mycomm.multiply.shift   = first1 + z2->first;
    mycomm.multiply.length1 = ((iindex==num-1) ? (z1->length-first1) : len1)+1;
    mycomm.multiply.length2 = z2->length - z2->first + 1;
    mycomm.multiply.first1  = 0;
    mycomm.multiply.first2  = 0;

    pstat[iprocs[iindex]].status    = BUSY;
    pstat[iprocs[iindex]].aufgabe   = MULT;
    pstat[iprocs[iindex]].bucket    = mycomm.multiply.bucket;
    pstat[iprocs[iindex]].shift     = mycomm.multiply.shift;

    fail = 0;
    SEND (pstat[mycomm.prozessor].fildes, 
	  &mycomm, sizeof (comm), sizeof (long), fail=1);

    if (fail) {
      /* Do something! */
    }

    /*
     * send the two numbers
     */

    SEND (pstat[mycomm.prozessor].fildes,
	  z1->digits + first1,
	  mycomm.multiply.length1 * sizeof (short),
	  sizeof (short),
	  return 1);
    SEND (pstat[mycomm.prozessor].fildes,
	  z2->digits + z2->first,
	  mycomm.multiply.length2 * sizeof (short),
	  sizeof (short),
	  return 1);

    first1   += mycomm.multiply.length1;
    totspeed -= pstat[iprocs[iindex]].spdidx;
  }
  /*
   * Hope we're done here
   */
  return 0;
}

int seldistmult (int bucket, int oneproc, int iprocs, bigint *z1, bigint *z2)
{
  int is1longer;
  /*
   * multiply on one processor if there is no free bucket (shouldn't happen),
   * not enough idle processors, or if the task is too tiny to distribute
   */

  if (bucket == -1 || iprocs < 2 ||
      z1->length - z1->first + z2->length - z2->first < 2000 ||
      z1->length - z1->first < 400 ||
      z2->length - z2->first < 400)
    return sendmultcom (oneproc, z1, z2);

  is1longer = (z1->length-z1->first > z2->length-z2->first) ? 1 : 0;

  return distmult (bucket,
		   iprocs,
		   is1longer ? z1 : z2,
		   is1longer ? z2 : z1);
}

int beschaeftige (int proc)
{
  int index, iprocs, fidle, thisbucket;
  long anfang, ende;

  /*
   * Count idle processors
   */
  
  for (index=0, iprocs=0, fidle=-1; index<MAXSLAVES; index++) {
    if (pstat[index].active && pstat[index].status == IDLE && pstat[index].spdidx > 10) {
      iprocs++;
      if (fidle == -1)
	fidle = index;
    }
  }

  /*
   * Search free bucket
   */
  for (index=0,thisbucket=-1; index<NUMBUCKETS; index++) {
    if (bucket[index].inuse == 0) {
      thisbucket = index;
      break;
    }
  }
  
  if (proc < 0)
    proc = fidle;
  
  if (proc>=MAXSLAVES || proc < 0)
    return 1;
  
  /*
   * If we have something to multiply, do it.
   */
  if (domult) {
    if (seldistmult (thisbucket, proc, iprocs, &mult1, &mult2) != 0) {
      fprintf (stderr, "Fatal error %d.\n", __LINE__);
      return 1;
    }
    
    domult = 0;
    mult1f = 0;
    mult2f = 0;
    return 0;
  }
  
  /*
   * If there are no multiplies to be done, send another slice.
   */
  if (sofar < theend) {
    anfang = sofar + 1;
    ende   = (sofar + step > theend) ? theend : sofar + step;
    
    if (sendaufgabe (proc, anfang, ende) != 0)
      return 1;
    
    sofar  = ende;
    return 0;
  }
  
  /*
   * Nothing left to do
   */
  return -1;
}

int calcloop (void)
{
  int res, busycount=0, flag=0, index, ctest;
  time_t ctime;
  comm mycomm;

  while (sofar < theend || busycount > 0 || flag == 0) {
    res   = event ();
    ctest = 0;
    RECV (pstat[res].fildes, &mycomm, sizeof (comm), sizeof (long), ctest=1);

    if (ctest == 1) {
      /*
       * Do something!
       */
      close (pstat[res].fildes);
      pstat[res].status = DEAD;
      continue;
    }

    flag = 1;

    switch (mycomm.reply) {
    case IDLE:
      pstat[res].status = IDLE;

      if (mycomm.speed > 0) {
	if (spiref == 0)
	  spiref = mycomm.speed;
	
	pstat[res].spdidx = (1000L*spiref)/mycomm.speed;
      }
      else 
        pstat[res].spdidx = 0;

      fprintf (stderr, "%8ld speed index of %d is %ld\n",
               AKTTIME(mtime), res, pstat[res].spdidx);
      break;
      
    case RESULT: /* result is coming in */
      if (mult1f == 1 && mult2f == 1) {
	fprintf (stderr, "%8ld Fatal Error! Result coming in and buffers full!\n", AKTTIME(mtime));
	bi_null (&mult1);
	bi_null (&mult2);
	
	mult1f  = 0;
	mult2f  = 0;
	domult  = 0;
      }
      
      ctime = time (NULL);
      if (mycomm.multiply.bucket != -1) {
	if (bucket[mycomm.multiply.bucket].inuse == 0)
	  fprintf (stderr, "%8ld Fatal error: got result from %02d for unused bucket %02ld\n",
		   AKTTIME(mtime), res, mycomm.multiply.bucket);
	if (bucket[mycomm.multiply.bucket].aktprocs == 0) {
	  bi_null (&bucket[mycomm.multiply.bucket].value);
	  RECV (pstat[res].fildes,
		bucket[mycomm.multiply.bucket].value.digits + mycomm.ergebnis.first,
		mycomm.ergebnis.length * sizeof (unsigned short),
		sizeof (short),
		return 1);
	  bucket[mycomm.multiply.bucket].value.first  = mycomm.ergebnis.first;
	  bucket[mycomm.multiply.bucket].value.length = mycomm.ergebnis.length + mycomm.ergebnis.first - 1;
	  bucket[mycomm.multiply.bucket].aktprocs    += 1;
	  
	  if (mycomm.multiply.shift > 0)
	    bi_shift_left (&bucket[mycomm.multiply.bucket].value,
			   mycomm.multiply.shift * 4L);
	}
	else {
	  bi_null (&rectemp);
	  RECV (pstat[res].fildes,
		rectemp.digits + mycomm.ergebnis.first,
		mycomm.ergebnis.length * sizeof (unsigned short),
		sizeof (short),
		return 1);
	  rectemp.first  = mycomm.ergebnis.first;
	  rectemp.length = mycomm.ergebnis.length + mycomm.ergebnis.first - 1;
	  
	  if (mycomm.multiply.shift > 0)
	    bi_shift_left (&rectemp,
			   mycomm.multiply.shift * 4L);
	  
	  bi_add (&bucket[mycomm.multiply.bucket].value,
		  &rectemp);
	  
	  bucket[mycomm.multiply.bucket].aktprocs += 1;
	  /*
	   * Bucket full?
	   */
	  if (bucket[mycomm.multiply.bucket].aktprocs ==
	      bucket[mycomm.multiply.bucket].numprocs) {
	    if (mult1f == 0) {
	      bi_assign (&mult1,
			 &bucket[mycomm.multiply.bucket].value);
	      mult1f = 1;
	    }
	    else /* if (mult2f == 0) */ {
	      bi_assign (&mult2,
			 &bucket[mycomm.multiply.bucket].value);
	      mult2f = 1;
	    }
	    bucket[mycomm.multiply.bucket].inuse = 0;
	  }
	}
      }
      else if (mult1f == 0) {
	bi_null (&mult1);
	RECV (pstat[res].fildes,
	      mult1.digits + mycomm.ergebnis.first, 
	      mycomm.ergebnis.length * sizeof (unsigned short),
	      sizeof (short),
	      return 1);
	mult1.length = mycomm.ergebnis.length + mycomm.ergebnis.first - 1;
	mult1.first  = mycomm.ergebnis.first;
	mult1f       = 1;
	
	if (mycomm.multiply.shift > 0)
	  bi_shift_left (&mult1,
			 mycomm.multiply.shift * 4L);
      }
      else /* if (mult2f == 0) */ {
	bi_null (&mult2);
	RECV (pstat[res].fildes,
	      mult2.digits + mycomm.ergebnis.first,
	      mycomm.ergebnis.length * sizeof (unsigned short),
	      sizeof (short),
	      return 1);
	mult2.length = mycomm.ergebnis.length + mycomm.ergebnis.first - 1;
	mult2.first  = mycomm.ergebnis.first;
	mult2f       = 1;
	
	if (mycomm.multiply.shift > 0)
	  bi_shift_left (&mult2,
			 mycomm.multiply.shift * 4L);
      }
      if (mult1f == 1 && mult2f == 1) 
	domult = 1;
      
      /*
       * give him credit
       */
      
      if (pstat[res].aufgabe == AUFGABE) {
	pstat[res].total     += mycomm.aufgabe.ende-mycomm.aufgabe.anfang+1;
	fprintf (stderr, "%8ld Got result of %6ld to %6ld from %02d  after %5ld secs\n",
		 AKTTIME(mtime),
		 mycomm.aufgabe.anfang,
		 mycomm.aufgabe.ende,
		 res,
		 mycomm.ergebnis.calctime);
      }
      else /* if (pstat[res].aufgabe == MULT) */ {
	pstat[res].mults ++;
	fprintf (stderr, "%8ld Got result of mult    %8ld from %02d  after %5ld secs\n",
		 AKTTIME(mtime),
		 mycomm.ergebnis.length - 1,
		 res,
		 mycomm.ergebnis.calctime);
      }
      pstat[res].calctime += mycomm.ergebnis.calctime;
      pstat[res].waittime += mycomm.ergebnis.waittime;
      pstat[res].transtime+= time(NULL) - ctime;
      
      /*
       * Declare him idle
       */
      pstat[res].status    = IDLE;
      pstat[res].aufgabe   = 0;
      pstat[res].anfang    = 0;
      pstat[res].ende      = 0;
      pstat[res].bucket    =-1;
      pstat[res].shift     = 0;

      /*
       * send new task
       */
      if (beschaeftige (res) == 0)
	flag--;
      break;
    } /* switch */
    
    /*
     * pass out some work
     */
    
    res = 0;
    while (flag > 0 && res == 0) {
      if ((res = beschaeftige (-1)) == 0)
	flag--;
    }
    
    /*
     * Count busy processors. We don't want to quit until they're
     * finished
     */
    for (index=0, busycount=0; index < MAXSLAVES; index++) {
      if (pstat[index].active && 
	  (pstat[index].status == BUSY ||
	   pstat[index].anfang != 0    ||
	   pstat[index].ende   != 0))
	busycount++;
    }
  } /* main while */
  
  /*
   * hey, it looks like we're finished. if we are, the result should
   * be in mult1 or mult2
   */
  if (domult || (mult1f == 1 && mult2f == 1)) {
    fprintf (stderr, "%8ld Fatal Error in master. Why are we finished if we aren't?\n", AKTTIME(mtime));
    return 1;
  }
  if (mult1f == 1) {
    bi_assign (&result, &mult1);
    mult1f = 0;
  }
  else if (mult2f == 1) {
    bi_assign (&result, &mult2);
    mult2f = 0;
  }
  else {
    fprintf (stderr, "%8ld Fatal Error in master. We're finished but don't have a result.\n", AKTTIME(mtime));
    return 1;
  }
  
  return 0;
}

int main (int argc, char *argv[])
{
  static char hntemp[256];
  struct hostent *he;
  int res, index;
  time_t c_ende;
  char *ptr;

  fprintf (stderr, "\nMP-Factorials\n\n");

  init   ();
  signal (SIGINT, quitonsig);
  signal (SIGTERM,quitonsig);

  if (getenv ("FACTORIAL") && getenv ("SLICE")) {
    if (argc >= 2)
      theend = atol (argv[1]);
    else
      theend = atol (getenv ("FACTORIAL"));

    step   = atol (getenv ("SLICE"));
  }
  else if (argc >= 2) {
    theend = atol (argv[1]);

    if (argc >= 3)
      step = atol (argv[2]);
    else if (getenv ("SLICE"))
      step = atol (getenv ("SLICE"));
    else
      step = theend / 32;
  }
  else {
    fprintf (stderr, "Factorial of ....  "); fflush (stdout); fflush (stderr); scanf ("%ld", &theend);
    fprintf (stderr, "Size of Slice ...  "); fflush (stdout); fflush (stderr); scanf ("%ld", &step);
    fprintf (stderr, "\n");
  }
  mtime   = 0;
  res     = calcloop ();
  c_ende  = time     (NULL);

  if (res != 0)
    fprintf (stderr, "Oops! Something went wrong.\n");

  fprintf (stderr, "\nUtilization statistics:\n");
  fprintf (stderr, " Proc | host     | speed | calctime | waittime | transtime | domain \n");
  fprintf (stderr, "------+----------+-------+----------+----------+-----------+--------\n");
  for (index=0; index<MAXSLAVES; index++) {
    if (pstat[index].active) {
      if ((he=gethostbyaddr((char*)&pstat[index].addr.sin_addr, 4, AF_INET))){
	strcpy (hntemp, he->h_name);
	if ((ptr = strchr (hntemp, '.'))) {
	  *ptr++ = '\0';
	  if (strchr (ptr, '.'))
	    *(strchr(ptr, '.')) = '\0';
	}
      }
      fprintf (stderr, "  %2d  | %-8s | %5ld | %7ld  | %7ld  |  %7ld  | %s\n",
	       index,
	       (he) ? hntemp : "unknown",
	       pstat[index].spdidx,
	       pstat[index].calctime,
	       pstat[index].waittime,
	       pstat[index].transtime,
	       (he) ? ptr : "");
    }
  }
  fprintf (stderr, "\n");

  fprintf (stderr, "Quitting ...\n");
  finish  ();
  fprintf (stderr, "Quitting finished.\n\n");

  if (res == 0) {
    bi_output (&result, stdout);
    printf ("\nCalculation time: %8ld secs\n\n",
	    c_ende - mtime);
  }
  return 0;
}

