#include <sys/types.h>
#include <sys/socket.h>
#include <sys/time.h>
#include <netinet/in.h>
#include <netdb.h>
#include <stdio.h>
#include <time.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include <signal.h>
#include <errno.h>
#include "PsxIO.h"
#include "comm.h"
#include "bigint.h"

/*
 * Global variables
 */

int    thesock;
int    status    = IDLE;
int    debug     = 0;
comm   thisaufgabe;
bigint thisnumber;
bigint mult1, mult2;
time_t sltime;

FILE  *thelog;

#define AKTTIME(s)	((long) (time(NULL)-(s)))

#define SEND(o,x,l,e,c) { \
			int xyz; \
			xyz = PsxWrite (o, (char *) (x), l, e); \
			if (xyz != l) { \
			  fprintf (thelog, "%8ld Fatal error slave (%d): %ld supposed %ld errno=%d\n", \
				   AKTTIME(sltime), __LINE__, (long)xyz, (long)l, errno); \
			  c; \
			} \
		  };

#define RECV(i,x,l,e,c) { \
		        int xyz; \
			xyz = PsxRead (i, (char *) (x), l, e); \
			if (xyz != l) { \
			  fprintf (thelog, "%8ld Fatal error slave (%d): %ld supposed %ld errno=%d\n", \
				   AKTTIME(sltime), __LINE__, (long)xyz, (long)l, errno); \
			  c; \
			} \
		  };

/*
 * Calculate speed index
 */

#define TESTLIMIT	2000

long calcspeed (void)
{
  struct timeval tvs, tve;
  struct timezone stz;
  long index, elapsed;

  gettimeofday (&tvs, &stz);

  bi_one (&mult1);
  bi_one (&mult2);
  bi_one (&thisnumber);

  for (index=1; index<=TESTLIMIT; index+=2)
    bi_mult_long (&mult1, index);
  for (index=2; index<=TESTLIMIT; index+=2)
    bi_mult_long (&mult2, index);

  bi_multiply (&thisnumber, &mult1, &mult2);

  bi_one (&mult1);
  bi_one (&mult2);
  bi_one (&thisnumber);

  gettimeofday (&tve, &stz);

  if (tvs.tv_usec > tve.tv_usec) {
    tve.tv_usec += 1000000;
    tve.tv_sec  -= 1;
  }
  elapsed = ((long) (1000*(tve.tv_sec - tvs.tv_sec) +
		     (tve.tv_usec - tvs.tv_usec)/1000));

  return elapsed;
}


/*
 * Calculate something. Either a new slice, or a multiplication
 */

void docalc (void)
{
  time_t c_begin=0, c_ende=0;
  long iter;

  bi_one (&thisnumber);
  c_begin = time (NULL);

  if (thisaufgabe.command == AUFGABE) {
    fprintf (thelog, "%8ld Calculating   %6ld to %6ld  ... ",
	    AKTTIME(sltime), 
	    thisaufgabe.aufgabe.anfang,
	    thisaufgabe.aufgabe.ende);
    fflush (thelog);

    for (iter =thisaufgabe.aufgabe.anfang;
	 iter<=thisaufgabe.aufgabe.ende;
	 iter++)
      bi_mult_long (&thisnumber, iter);

    fprintf (thelog, "done after %5ld secs.\n", AKTTIME (c_begin));
  }
  else if (thisaufgabe.command == MULT) {
    fprintf (thelog, "%8ld Multiplying   %6ld x  %-6ld  ... ",
	    AKTTIME(sltime), mult1.length, mult2.length);
    fflush (thelog);
    
    bi_multiply (&thisnumber, &mult1, &mult2);

    fprintf (thelog, "done after %5ld secs.\n", AKTTIME (c_begin));
  }
  else {
    fprintf (thelog, "Illegal task in slave (%d): %ld\n",
	     __LINE__, thisaufgabe.command);
  }
  c_ende = time (NULL);

  thisaufgabe.ergebnis.length   = thisnumber.length - thisnumber.first + 1;
  thisaufgabe.ergebnis.first    = thisnumber.first;
  thisaufgabe.ergebnis.calctime = c_ende - c_begin;

  if (status == BUSY) /* might be terminate now */ {
    status = DONE;
  }
}

void loop (void)
{
  time_t w_begin, w_ende;
  int ok=1, flag;
  comm mycomm;

  fprintf (thelog, "%8ld Slave is waiting ... \n", AKTTIME(sltime));
  while (ok) {
    w_begin = time(NULL);
    RECV (thesock, &mycomm, sizeof (comm), sizeof (long), return);
    w_ende  = time(NULL);

    switch (mycomm.command) {
    case QUIT:
      fprintf (thelog, "%8ld Quitting ... \n", AKTTIME(sltime));
      ok = 0;
      break;
    case AUFGABE:				/* Aufgabe  */
      memcpy (&thisaufgabe, (char *) &mycomm, sizeof (comm));
      status = BUSY;
      docalc ();
      break;
    case MULT:
      bi_null (&mult1);
      bi_null (&mult2);

      RECV (thesock,
	    mult1.digits + mycomm.multiply.first1,
	    mycomm.multiply.length1 * sizeof (short),
	    sizeof (short),
	    ;);
      RECV (thesock,
	    mult2.digits + mycomm.multiply.first2,
	    mycomm.multiply.length2 * sizeof (short),
	    sizeof (short),
	    ;);

      mult1.length = mycomm.multiply.length1 + mycomm.multiply.first1 - 1;
      mult2.length = mycomm.multiply.length2 + mycomm.multiply.first2 - 1;
      mult1.first  = mycomm.multiply.first1;
      mult2.first  = mycomm.multiply.first2;
      
      memcpy (&thisaufgabe, (char *) &mycomm, sizeof (comm));
      status = BUSY;
      docalc ();
      break;
    case STATUS:
      mycomm.reply = status;
      mycomm.speed = calcspeed ();
      SEND (thesock, &mycomm, sizeof (comm), sizeof (long), ;);
      break;
    default:
      fprintf (thelog, "Illegal command slave (%d):%ld\n",
	       __LINE__, mycomm.command);
      return;
      break;
    }
    /*
     * Check if we have finished our work
     */
    if (status == DONE) {
      memcpy ((char *) &mycomm, &thisaufgabe, sizeof (comm));
      mycomm.ergebnis.waittime = w_ende - w_begin;
      mycomm.command           = NONE;
      mycomm.reply             = RESULT;

      /*
       * Test if connection still exists
       */
      flag = 0;
      RECV (thesock, NULL, 0, 1, flag=1);
      if (flag == 1) {
	fprintf (thelog, "Error! Master has closed connection.\n");
	return;
      }

      SEND (thesock, &mycomm, sizeof (comm), sizeof (long), ;);
      SEND (thesock,
	    thisnumber.digits + thisnumber.first,
	    mycomm.ergebnis.length * sizeof (short),
	    sizeof (short),
	    ;);

      status = IDLE;				/* Go idle again */
    }
  }
}

int main (int argc, char *argv[])
{
  struct hostent *theent;
  struct sockaddr_in theaddr;
  long thespeed;
  comm mycomm;

  /*
   * go background
   */

  if (fork ()) {
    exit (0);
  }

  close (0);
  close (1);
  close (2);

  if ((thelog = fopen ("/dev/null", "w")) == NULL)
    exit (0);
  
  bi_const (&thisnumber);
  bi_const (&mult1);
  bi_const (&mult2);

  /*
   * get speed 
   */
  fprintf  (thelog, "Getting speed index ... ");
  fflush   (thelog);
  thespeed = calcspeed ();
  fprintf  (thelog, "%ld.\n", thespeed);

  if (thespeed == 0) {
    fprintf (thelog, "Fatal error: Speed is zero. Exiting.\n");
    fclose  (thelog);
    exit    (1);
  }

  if ((theent = gethostbyname (argv[1])) == NULL) {
    fprintf (thelog, "Error: couldn't lookup %s (%d).\n", argv[1], errno);
    fclose  (thelog);
    exit    (1);
  }

  theaddr.sin_family      = AF_INET;
  theaddr.sin_port        = htons (PORTNUMBER);
  theaddr.sin_addr.s_addr = *((long *)theent->h_addr);

  if ((thesock = socket (AF_INET, SOCK_STREAM, 0)) == -1) {
    fprintf (thelog, "Error: Couldn't create socket (%d).\n", errno);
    fclose  (thelog);
    exit    (1);
  }

  if (connect (thesock, (struct sockaddr *) &theaddr, sizeof (struct sockaddr_in)) == -1) {
    fprintf (thelog, "Error: Couldn't connect to server (%d).\n", errno);
    fclose  (thelog);
    close   (thesock);
  }

  /*
   * Hey, we're here!
   */

  sltime       = time(NULL);
  mycomm.reply = IDLE;
  mycomm.speed = thespeed;

  SEND   (thesock, &mycomm, sizeof (comm), sizeof (long), 
	  close(thesock);exit(1));
  loop   ();
  close  (thesock);

  bi_dest (&thisnumber);
  bi_dest (&mult1);
  bi_dest (&mult2);
  fclose  (thelog);

  return 0;
}
